function ConformalMapper(cornerBox, c, varargin)
%% Conformal mapper
% 
% Given a polygon (cornerBox) represented as an array of complex points 
% and a chosen conformal center, a conformal map is constructed that maps 
% the domain to the unit disk


cornerPolygon = polygon(cornerBox);

lims = boundingbox(cornerPolygon);
x0 = lims(1)-0.1;
x1 = lims(2)+0.1;
y0 = lims(3)-0.1;
y1 = lims(4)+0.1;
x = x0:0.0999999:x1;
y = y0:0.0999999:y1;

fishNet = {};
for t = x
    [xi, yi] = polyxpoly([t, t], [y0 y1], real(cornerBox), imag(cornerBox));
    if size(xi,1) == 0
        continue
    end
    T = [xi yi];
    T = sortrows(T, 2);
    numSegments = int32(size(T,1)/2);
    if(mod(size(T,1),2) == 0)
        pairs = reshape(T', [2, 2, numSegments]);
        for it = 1:numSegments
            M = pairs(:,:,it)';
            C = M(:,1)+M(:,2)*1j;
            AB = C(2)-C(1);
            R = C(1) + [0:0.005:1]*AB;
            fishNet(end + 1) = {R};
        end
    end
end
    
for t = y
    [xi, yi] = polyxpoly([x0 x1],[t, t] , real(cornerBox), imag(cornerBox));
    if size(xi,1) == 0
        continue
    end
    T = [xi yi];
    T = sortrows(T, 1);
    numSegments = int32(size(T,1)/2);
    if(mod(size(T,1),2) == 0)
        pairs = reshape(T', [2, 2, numSegments]);
        for it = 1:numSegments
            M = pairs(:,:,it)';
            C = M(:,1)+M(:,2)*1j;
            AB = C(2)-C(1);
            R = C(1) + [0:0.005:1]*AB;
            fishNet(end + 1) = {R};
        end
    end
end

mapping = crdiskmap(cornerPolygon);
mapping = center(mapping, c);

mapped = {};
for i = 1:size(fishNet,2)
    mapped(end + 1) = {evalinv(mapping,fishNet{i})};
end


%% Density map
[X, Y] = meshgrid(x0:0.0099999:x1, y0:0.0099999:y1);

X = reshape(X,[],1)';
Y = reshape(Y,[],1)';
originalPoints = inpolygon(X, Y, real(cornerBox),imag(cornerBox));

C = X + 1j*Y;
originalPoints = C(originalPoints);

virtualPoints = evalinv(mapping, originalPoints);
magnitudes = abs(virtualPoints);
radialReflection = virtualPoints(magnitudes>0.1);
magnitudes = magnitudes(magnitudes>0.1);
radialReflection = (2-magnitudes).*radialReflection./(magnitudes);

extendedPoints = horzcat(virtualPoints, radialReflection);
xCoords = real(extendedPoints);
xMax = max(xCoords);
xMin = min(xCoords);

yCoords = imag(extendedPoints);
yMax = max(yCoords);
yMin = min(yCoords);

x = xMin:0.01:xMax;
y = yMin:0.01:yMax;
[X, Y] = meshgrid(x, y);

pts = [X(:) Y(:)];
coords = [xCoords' yCoords'];

f = ksdensity(coords, pts, 'Bandwidth', 0.05);
f = reshape(f, size(X));
averageResolution = pi/length(virtualPoints);
f = f - averageResolution;

% Map the densitymap back to the original domain
u = x0:0.0099999:x1;
v = y0:0.0099999:y1;
originalDensity = NaN(length(u), length(v));

for it = 1:length(originalPoints)
   oz = originalPoints(it);
   ozx = ceil((real(oz) - x0)*length(u)/(x1-x0));
   ozy = ceil((imag(oz) - y0)*length(v)/(y1-y0));
   
   vz = virtualPoints(it);
   vzx = ceil((real(vz)/2 + 1)*length(x)/2);
   vzy = ceil((imag(vz)/2 + 1)*length(y)/2);
   
   originalDensity(ozx, ozy) = f(vzy,vzx);
end



%% Pathfinding
computePath = false;

if length(varargin) == 3
    
   computePath = true; 
    
   p1 = cell2mat(varargin(1));
   p2 = cell2mat(varargin(2));
   
   v1 = evalinv(mapping, p1);
   v2 = evalinv(mapping, p2);
   
   vPath = linspace(v1, v2, 1000);
   oPath = mapping(vPath);
   
   originalBestPath = cell2mat(varargin(3));
   orginalBestPathVirtual = [];
   for it = 2:length(originalBestPath)
       orginalBestPathVirtual = [orginalBestPathVirtual evalinv(mapping, linspace(originalBestPath(it-1),originalBestPath(it), 1000))]; 
   end
    
end



%% Plotting

figure;
hold on;
plot(cornerBox, 'Color', [0.2578    0.5234    0.9531], 'LineWidth', 1);
for i = 1:size(fishNet,2)
    plot(fishNet{i},'k');
end

if computePath
    plot(oPath, 'r', 'LineWidth', 1);
    plot(originalBestPath, 'b','LineWidth', 1);
end
bounds = boundingbox(cornerPolygon) + [-0.5 0.5 -0.5 0.5];
axis(bounds);
axis equal;
axis off;

figure;
hold on; 
axis([-1.5 1.5 -1.5 1.5]);

plot(exp(1i*linspace(0,2*pi,180)), 'Color', [0.2578    0.5234    0.9531], 'LineWidth', 1);
for i = 1:size(mapped,2)
    plot(mapped{i},'k');
end
if computePath
   plot([v1 v2], 'r','LineWidth', 1); 
   plot(orginalBestPathVirtual, 'b','LineWidth', 1);
end
axis equal;
axis off;
hold off;

figure;
im = image(originalDensity','CDataMapping','scaled');
set(gca,'YDir','Normal');
im.AlphaData = ~isnan(originalDensity');
axis equal;
axis off;


figure;
m = X.^2+Y.^2<1;
im = image(f,'CDataMapping','scaled');
set(gca,'YDir','Normal');
im.AlphaData = m;
axis equal;
axis off;
imwrite(f,'heatmap.png');

